#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

//#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>

//Mode0
//BG3 +a

//---------------------------------------------------------------------------
//is
#define STATE_INIT 				1
#define STATE_RUN  				3
#define STATE_END				4

//\pBG^C
#define BG_TILE_EMPTY			0
#define BG_TILE_FOOD			1
#define BG_TILE_SNAKE			2

#define SNAKE_MAX_TAIL_CNT		50
#define SNAKE_MOVE_TIME			200

#define SRAM_SRAND_OFFSET		0

//---------------------------------------------------------------------------
typedef struct {
	u8  x;
	u8  y;
} ST_TAIL;

typedef struct {
	u8  x;
	u8  y;
} ST_FOOD;

typedef struct {
	u8      x;
	u8      y;
	s16     minX;
	s16     maxX;
	s16     minY;
	s16     maxY;
	s16     moveX;
	s16     moveY;
	u8      key;
	u8      tailCnt;
	ST_TAIL tail[SNAKE_MAX_TAIL_CNT];
} ST_SNAKE;

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_BG    BG3;
ST_SNAKE Snake;
ST_FOOD  Food;

u8       State;
u32      SnakeMoveTime;

//---------------------------------------------------------------------------
void StateInit();
void StateRun();
void StateEnd();

void InitWS();
void InitBG();
void InitKey();
void IRQUserHandler();
void InitSrand();
void InitSnake();

void ProcKeyInput();

void MoveSnake();
void DrawSnake(bool isDraw);
bool ChkHitSnake();
bool ChkHitWall();
bool ChkHitTail();

void SetFood();
void DrawFood(bool isDraw);
bool ChkHitFood();

void SramSaveSrand();

//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:  StateInit();  break;
			case STATE_RUN:   StateRun();   break;
			case STATE_END:   StateEnd();   break;
			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitBG();
	InitKey();
	InitSrand();
	InitSnake();

	SetFood();

	DrawSnake(TRUE);
	DrawFood(TRUE);

	SnakeMoveTime = 0;

	SetMode(MODE_0 | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	State = STATE_RUN;
}
//---------------------------------------------------------------------------
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG3 +a

	BG3.chr = 1;
	BG3.map = 0;
	BG3.col = 0;

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
	              BG_CHARBASE(BG3.chr) | BG_MAPBASE(BG3.map) | 0);

	BGSetPal16  (BG3.col, (u16*)&game_bg_pal_bin);
	BGSetChrData(BG3.chr, (u16*)&game_bg_img_bin, BG_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();

	//L[͗p̃^C}ݒ
	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;

	REG_TM0D     = 65536 - 2621; // (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	SnakeMoveTime   += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void InitSnake()
{
	ST_SNAKE* p = &Snake;
	u8 def_x    = (240 / 8) / 2;
	u8 def_y    = (160 / 8) / 2;

	p->x        = def_x;
	p->y        = def_y;
	p->minX     = 0;
	p->minY     = 0;
	p->maxX     = 29;
	p->maxY     = 19;
	p->moveX    = 1;
	p->moveY    = 1;
	p->key      = 0x00;

	s16 i;
	for(i=0; i<SNAKE_MAX_TAIL_CNT; i++)
	{
		p->tail[i].x = def_x;
		p->tail[i].y = def_y;
	}
	p->tailCnt = 0;
}
//---------------------------------------------------------------------------
void StateRun()
{
	ProcKeyInput();

	if(SnakeMoveTime >= SNAKE_MOVE_TIME)
	{
		SnakeMoveTime = 0;

		if( ChkHitSnake() == TRUE ) //ʊOAKɓ`FbN܂
		{
			SramSaveSrand();
			State = STATE_END;
			return;
		}
		if( ChkHitFood() == TRUE ) //aɓ`FbN܂
		{
			Snake.tailCnt = (Snake.tailCnt + 1) % SNAKE_MAX_TAIL_CNT;

			SetFood();
			DrawFood(TRUE);
			return;
		}

		DrawSnake(FALSE);
		MoveSnake();
		DrawSnake(TRUE);
	}
}
//---------------------------------------------------------------------------
void ProcKeyInput()
{
	if( KeyWaitTimer() == TRUE )
	{
		return;
	}
	if( KeyRepeat() == TRUE )		//keyGet֐ĂԑOɕKsȂ
	{
		return;
	}

	u8 key = KeyGet();
	if(key == 0x00)
	{
		return;
	}

	ST_SNAKE* p   = &Snake;
	p->key        = key;
	SnakeMoveTime = SNAKE_MOVE_TIME;

	rand();							//Q[Jnɗ𔭐
}
//---------------------------------------------------------------------------
bool ChkHitSnake()
{
	bool isRet = FALSE;

	isRet |= ChkHitWall();
	isRet |= ChkHitTail();

	return isRet;
}
//---------------------------------------------------------------------------
bool ChkHitWall()
{
	ST_SNAKE* p = &Snake;

	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->key & KEY_UP   ) ty1 -= p->moveY;
	if(p->key & KEY_DOWN ) ty1 += p->moveY;
	if(p->key & KEY_LEFT ) tx1 -= p->moveX;
	if(p->key & KEY_RIGHT) tx1 += p->moveX;

	if(tx1 < p->minX || ty1 < p->minY) return TRUE;
	if(tx1 > p->maxX || ty1 > p->maxY) return TRUE;

	return FALSE;
}
//---------------------------------------------------------------------------
bool ChkHitTail()
{
	ST_SNAKE* p = &Snake;

	u8 x = p->x;
	u8 y = p->y;

	s16 i;
	for(i=0; i<p->tailCnt; i++)
	{
		if(x == p->tail[i].x && y == p->tail[i].y)
		{
			return TRUE;
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
void MoveSnake()
{
	ST_SNAKE* p = &Snake;

	//Tail
	s16 i;
	for(i=p->tailCnt-1; i>=1; i--)
	{
		p->tail[i].x = p->tail[i-1].x;
		p->tail[i].y = p->tail[i-1].y;
	}
	p->tail[0].x = p->x;
	p->tail[0].y = p->y;

	//Head
	if(p->key & KEY_UP   ) p->y -= p->moveY;
	if(p->key & KEY_DOWN ) p->y += p->moveY;
	if(p->key & KEY_LEFT ) p->x -= p->moveX;
	if(p->key & KEY_RIGHT) p->x += p->moveX;
}
//---------------------------------------------------------------------------
void StateEnd()
{
	
}
//---------------------------------------------------------------------------
void DrawSnake(bool isDraw)
{
	u8 tileNo;
	if(isDraw) tileNo = BG_TILE_SNAKE;
	else       tileNo = BG_TILE_EMPTY;

	ST_SNAKE* p = &Snake;

	//Head
	BGDrawChar(BG3.map, p->x, p->y, BG3.col, tileNo);

	//Tail
	s16 i;
	for(i=0; i<p->tailCnt; i++)
	{
		BGDrawChar(BG3.map, p->tail[i].x, p->tail[i].y, BG3.col, tileNo);
	}
}
//---------------------------------------------------------------------------
void InitSrand()
{
	u32 r = SRAMRead32(SRAM_SRAND_OFFSET);
	srand(r);
}
//---------------------------------------------------------------------------
void SetFood()
{
	ST_FOOD* p = &Food;

	do {
		p->x = rand() % (BG_SCREEN_CX - 1);
		p->y = rand() % (BG_SCREEN_CY - 1);

	} while( ChkHitFood() == TRUE );
}
//---------------------------------------------------------------------------
void SramSaveSrand()
{
	SRAMWrite32(SRAM_SRAND_OFFSET, (u32)rand() );
}
//---------------------------------------------------------------------------
void DrawFood(bool isDraw)
{
	u8 tileNo;
	if(isDraw) tileNo = BG_TILE_FOOD;
	else       tileNo = BG_TILE_EMPTY;

	ST_FOOD* p = &Food;
	BGDrawChar(BG3.map, p->x, p->y, BG3.col, tileNo);
}
//---------------------------------------------------------------------------
bool ChkHitFood()
{
	ST_FOOD*  f = &Food;
	ST_SNAKE* s = &Snake;

	u8 x = f->x;
	u8 y = f->y;

	//Head
	if(x == s->x && y == s->y)
	{
		return TRUE;
	}

	//Tail
	s16 i;
	for(i=0; i<s->tailCnt; i++)
	{
		if(x == s->tail[i].x && y == s->tail[i].y)
		{
			return TRUE;
		}
	}

	return FALSE;
}
